/* Copyright: Jonathan Wright, 2001*/

/*          ****************
            *FILE:COMPLEX.C*
            ****************        */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "complex.h"

struct complex_method cpx;

/*--------------------------------------------------------
Action:"Complex Method" Heuristic Search Algorithm by Box.

Author: J.A.Wright
Date:5-8-00
----------------------------------------------------------*/
void box(void){

double alpha,f_reflect;
int    i_vert,i_var,index_worst;
int    order[(2*MAX_SEARCH_VAR+2)];
int    cont,cont_reflect;

/*Notes:
1. all variables operated on in a normalized form of 0 to 1;
2. the number of vertices is fixed at (2*number_variables+2);
3. initial guess of solution, number of variables, and variable
   bounds must be set in the calling program through the
   structure "cpx".
4. objective function evaluated through calc_objective();
5. constraint functions evaluated through test_implicit();*/

/*Check the number of variables is not greater than MAX_SEARCH_VAR*/

    if (cpx.n_var>MAX_SEARCH_VAR) {
        printf("Number of variables in the Complex search exceeds the maximum of %d\n",MAX_SEARCH_VAR);
        printf("Reduce the number or increase MAX_SEARCH_VAR and recompile!\n");
        printf("Press any key to exit.\n");getchar();exit(0);
    }

/*Initilialize variables*/

    cpx.n_func_calls=0;
	initialize_box();

/*Evaluate the objective function values for the initial complex*/

	for (i_vert=0;i_vert<(2*cpx.n_var+2);i_vert++)
	                        cpx.f[i_vert] = calc_objective(cpx.vertex[i_vert]);

/*Reflect worst point until "complex" converges*/

	do {
		/*Get an index for the order of objective function values*/

		obj_sort(cpx.f,(2*cpx.n_var+2),order);
		index_worst=(2*cpx.n_var+2)-1; /*Highest value==last in index "order"*/

        /*Reflect worst solution about centroid of remaining vetrices*/

		do {
    		alpha = ALPHA;

		    /*Copy variable values of "worst" solution to worst[]*/

		    copy_worst(order[index_worst]);

		    /*Calculate centroid of complex, excluding the worst solution*/

			calc_centroid(order[index_worst],(2*cpx.n_var+2));

			/*Reflect the worst solution about the centroid*/

			reflect(order[index_worst],alpha);

            /*Calculate objective function value of reflected point*/

            f_reflect=calc_objective(cpx.vertex[order[index_worst]]);

            /*Print reflected point if monit set*/

            if (cpx.monit==MONIT) {
                printf("norm(X[]): ");
	            for(i_var=0;i_var<cpx.n_var;i_var++)
	                printf("%.2f ",cpx.vertex[order[index_worst]][i_var]);
	                printf("f(X[]): %lf\n",f_reflect);
	        }

            /*Check that the reflected objective better than
              the unreflected objective, and if not, reduce alpha
              until alpha less than tollerance*/

			while ((f_reflect>=cpx.f[order[index_worst]]) &&
			        (alpha>LOWEST_ALPHA)) {
				alpha /= 2.0;
			    reflect(order[index_worst],alpha);
                f_reflect=calc_objective(cpx.vertex[order[index_worst]]);
			}

            /*Re-check that the reflected objective better than
              the unreflected objective, and if not, reset
              the worst solution vertex to the unreflected values
              and start reflection using the next worst solution*/

            if (f_reflect>=cpx.f[order[index_worst]]) {
                reset_worst(order[index_worst]);
                --index_worst; /*Index for next worst*/
                if (index_worst<0) {/*All vertices tried, stop search*/
                    cont=STOP;
                    cont_reflect=STOP;
                }
                else {/*More vertices to try, continue reflection process*/
                    cont=CONTINUE;
                    cont_reflect=CONTINUE;
                }
            }

            /*Reflected point better than unreflected point,
              set new objective function value and set a flag
              to break out of the loop*/

            else {
                cont=STOP;
                cont_reflect=CONTINUE;
            }

		} while (cont);

        cpx.f[order[index_worst]]=calc_objective(cpx.vertex[order[index_worst]]);


	} while ((cont_reflect) && (!converge()));

/*Search converged or stoped, copy best solution to optimum[]*/

	obj_sort(cpx.f,(2*cpx.n_var+2),order);
	copy_optimum(order[0]);

    /*Print optimum point if monit set*/

    if (cpx.monit==MONIT) {
        printf("Optimum(X[]): ");
	    for(i_var=0;i_var<cpx.n_var;i_var++)
	            printf("%.2f ",cpx.optimum[i_var]);
	    printf("\nOptimum_f(X[]): %lf\nNumber f(X[]) calls: %d\n",cpx.f_optimum,cpx.n_func_calls);
	    printf("Complex search complete, press any key to continue.\n");
	    getchar();
	}

}

/*---------------------------------------------------------
Action: Set variable values for the initial "complex".

Author: J.A.Wright
Date:5-8-00
-----------------------------------------------------------*/
void initialize_box(void){
int i_var,i_vert;

/*Set first vertex to the initial feasible guess*/

	for (i_var=0;i_var<cpx.n_var;i_var++) {
		cpx.vertex[0][i_var]=(cpx.x_guess[i_var]-cpx.x_lb[i_var])/(cpx.x_ub[i_var]-cpx.x_lb[i_var]);
	}

/*Check feasibility of initial guess*/

    if (test_constraints(cpx.vertex[0])) {
        printf("The initial guess of the solution in the 'complex' search\n");
        printf("is infeasible; check constraint formulation and guess.\n");
        printf("Press any key to exit.\n");getchar();exit(0);
    }

/*Initialize the remaining vertices of "complex"*/

	for (i_vert=1;i_vert<(2*cpx.n_var+2);i_vert++) {

	    /*Generate variable value for current vertex*/

		for (i_var=0;i_var<cpx.n_var;i_var++) cpx.vertex[i_vert][i_var] = jrandom();

        /*Calculate the  centroid of the "complex", excluding current vertex*/

        calc_centroid(i_vert,(i_vert+1));

		/*Test feasibility of the current vertex and correct if neccesary*/

		test_and_move(i_vert);
	}

}

/*--------------------------------------------
Action: Calculate the centroid of the complex,
        excluding the worst solution.

Author: J.A.Wright
Date:7-8-00
----------------------------------------------*/
void calc_centroid(int i_worst,int n_vert){
int i_var,i_vert;

/*Calculate the centroid of complex, excluding i_worst vertex*/

	for (i_var=0;i_var<cpx.n_var;i_var++) {
		cpx.centroid[i_var] = 0.0;
		for (i_vert=0;i_vert<(2*cpx.n_var+2);i_vert++) {
		    if (i_vert != i_worst) cpx.centroid[i_var] += cpx.vertex[i_vert][i_var];
        }
		cpx.centroid[i_var] /= (n_vert-1);
	}
}

/*-------------------------------------------
Action: Identify an index for the ascending
        order of objective function values,
        (maximum objective == last in index).

Author: J.A.Wright
Date:7-8-00
---------------------------------------------*/
void obj_sort(double objective[],int n_obj,int *sort){

int i,j,k;

    sort[0]=0;
    for (i=1;i<n_obj;i++){
        j=0;
        while(objective[sort[j]]<objective[i]) {
            ++j;if(j>(i-1))break;
        }
        for (k=i;k>=j;k--) sort[k]=sort[k-1];
        sort[j]=i;
    }
}

/*--------------------------------------------------------------
Action: Test the feasibility of a vertex and move if infeasible.

Author: J.A.Wright
Date:9-8-00
----------------------------------------------------------------*/
void test_and_move(int i_vertex){
int i_var;

/*Test bound constraints and move violations back inside the bounds*/

	for(i_var=0;i_var<cpx.n_var;i_var++) {
		if      (cpx.vertex[i_vertex][i_var] > 1.0) cpx.vertex[i_vertex][i_var]=0.9999999;
		else if (cpx.vertex[i_vertex][i_var] < 0.0) cpx.vertex[i_vertex][i_var]=0.0000001;
	}

/*Test constraint violations, and move infeasible points back towards the centroid*/

	while (test_constraints(cpx.vertex[i_vertex])) {
		for(i_var=0;i_var<cpx.n_var;i_var++)
		    cpx.vertex[i_vertex][i_var]=(cpx.centroid[i_var]+cpx.vertex[i_vertex][i_var])/2.0;
	}
}

/*-------------------------------------
Action:Copy the variable values for the
       "worst" vertex to worst[].

Author:J.A.Wright
Date:7-8-00
----------------------------------------*/
void copy_worst(int i_worst){

int i_var;

    for (i_var=0;i_var<(2*cpx.n_var+2);i_var++)
                cpx.worst[i_var]=cpx.vertex[i_worst][i_var];
}

/*------------------------------------------
Action: Reflect the worst solution about the
        centroid of the remaining points.

Author: J.A.Wright
Date:5-8-00
--------------------------------------------*/
void reflect(int i_worst, double alpha){
int i_var;

/*Reflect Worst Solution about the Centroid of the remaining vetrices*/

	for(i_var=0;i_var<cpx.n_var;i_var++) cpx.vertex[i_worst][i_var]=
	                                        ((1.0+alpha)*cpx.centroid[i_var])-
	                                           (alpha*cpx.vertex[i_worst][i_var]);

/*Check Feasibility of the Reflected Vertex and move point until feasible*/

	test_and_move(i_worst);

}

/*-------------------------------------
Action:Reset the variable values for the
       "worst" vertex.

Author:J.A.Wright
Date:7-8-00
----------------------------------------*/
void reset_worst(int i_worst){

int i_var;

    for (i_var=0;i_var<(2*cpx.n_var+2);i_var++)
                cpx.vertex[i_worst][i_var]=cpx.worst[i_var];
}

/*-----------------------------------------
Action: Test for Convergence of the search.

Author:J.A.Wright
Date:7-8-00
-------------------------------------------*/
int converge(void){

int converge,i;
double f_cent,std_dev;

    converge=!CONVERGE;

/*Check maximum number of objective function calls*/

    if (cpx.n_func_calls>=cpx.max_func_call) converge=CONVERGE;

/*Check standard deviation of objective function values about centroid */

    std_dev=0;
    f_cent=calc_objective(cpx.centroid);
    for (i=0;i<(2*cpx.n_var+2);i++)
        std_dev += sqrt(pow((f_cent-cpx.f[i]),2));
    std_dev /= (2*cpx.n_var+2);
    if (std_dev<=cpx.deviation_tollerance) converge=CONVERGE;

return converge;
}

/*-------------------------------------
Action: Copy the best solution to
       optimum[] and f_optimum.

Author:J.A.Wright
Date:7-8-00
----------------------------------------*/
void copy_optimum(int i_best){

int i_var;

/*Copy descaled optimum values to optimum[]*/

    for (i_var=0;i_var<(2*cpx.n_var+2);i_var++)
                cpx.optimum[i_var]=cpx.x_lb[i_var]+
                                    cpx.vertex[i_best][i_var]*
                                    (cpx.x_ub[i_var]-cpx.x_lb[i_var]);

/*Copy optimum objective function value to f_optimum*/

    cpx.f_optimum=cpx.f[i_best];
}

/*--------------------------------------------------
Action: Return a random number betwen 0.0 and 1.0.

Author: M.Matsumoto
Date: 5-8-00
----------------------------------------------------*/
double jrandom(void){

/* A C-program for TT800 : July 8th 1996 Version */
/* by M. Matsumoto, email: matumoto@math.keio.ac.jp */
/* genrand() generate one pseudorandom number with double precision */
/* which is uniformly distributed on [0,1]-interval */
/* for each call.  One may choose any initial 25 seeds */
/* except all zeros. */

/* See: ACM Transactions on Modelling and Computer Simulation, */
/* Vol. 4, No. 3, 1994, pages 254-266. */

#define N 25
#define M 7

    unsigned long y;
    static int k = 0;
    static unsigned long x[N]={ /* initial 25 seeds, change as you wish */
        0x95f24dab, 0x0b685215, 0xe76ccae7, 0xaf3ec239, 0x715fad23,
        0x24a590ad, 0x69e4b5ef, 0xbf456141, 0x96bc1b7b, 0xa7bdf825,
        0xc1de75b7, 0x8858a9c9, 0x2da87693, 0xb657f9dd, 0xffdc8a9f,
        0x8121da71, 0x8b823ecb, 0x885d05f5, 0x4e20cd47, 0x5a9ad5d9,
        0x512c0c03, 0xea857ccd, 0x4cc1d30f, 0x8891a8a1, 0xa6b7aadb
    };
    static unsigned long mag01[2]={
        0x0, 0x8ebfd028 /* this is magic vector `a', don't change */
    };
    if (k==N) { /* generate N words at one time */
      int kk;
      for (kk=0;kk<N-M;kk++) {
        x[kk] = x[kk+M] ^ (x[kk] >> 1) ^ mag01[x[kk] % 2];
      }
      for (; kk<N;kk++) {
        x[kk] = x[kk+(M-N)] ^ (x[kk] >> 1) ^ mag01[x[kk] % 2];
      }
      k=0;
    }
    y = x[k];
    y ^= (y << 7) & 0x2b5b2500; /* s and b, magic vectors */
    y ^= (y << 15) & 0xdb8b0000; /* t and c, magic vectors */
    y &= 0xffffffff; /* you may delete this line if word size = 32 */
/*
   the following line was added by Makoto Matsumoto in the 1996 version
   to improve lower bit's corellation.
   Delete this line to o use the code published in 1994.
*/
    y ^= (y >> 16); /* added to the 1994 version */
    k++;
    return( (double) y / (unsigned long) 0xffffffff);
}
